﻿Namespace NeuralNetworks
    Friend Module NeuralNetworksProgram
        Sub Main(ByVal args() As String)
            Try
                Console.WriteLine(vbCrLf & "Begin Neural Network demo" & vbCrLf)

                Console.WriteLine("Creating a 3-input, 4-hidden, 2-output neural network")
                Console.WriteLine("Using sigmoid function for input-to-hidden activation")
                Console.WriteLine("Using tanh function for hidden-to-output activation")
                Dim nn As New NeuralNetwork(3, 4, 2)

                Dim weights() As Double = {0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0, 1.1, 1.2, -2.0, -6.0, -1.0, -7.0, 1.3, 1.4, 1.5, 1.6, 1.7, 1.8, 1.9, 2.0, -2.5, -5.0}

                Console.WriteLine(vbCrLf & "Weights and biases are:")
                Helpers.ShowVector(weights)

                Console.WriteLine("Loading neural network weights and biases")
                nn.SetWeights(weights)

                Console.WriteLine(vbCrLf & "Setting inputs:")
                Dim xValues() As Double = {1.0, 2.0, 3.0}
                Helpers.ShowVector(xValues)

                Console.WriteLine("Loading inputs and computing outputs")
                Dim yValues() As Double = nn.ComputeOutputs(xValues)

                Console.WriteLine(vbCrLf & "Neural network outputs are:")
                Helpers.ShowVector(yValues)

                Console.WriteLine("End Neural Network demo" & vbCrLf)
                Console.ReadLine()
            Catch ex As Exception
                Console.WriteLine("Fatal: " & ex.Message)
                Console.ReadLine()
            End Try
        End Sub ' Main
    End Module  ' Module NeuralNetworksProgram

    Friend Class NeuralNetwork
        Private numInput As Integer
        Private numHidden As Integer
        Private numOutput As Integer

        Private inputs() As Double
        Private ihWeights()() As Double ' input-to-hidden
        Private ihSums() As Double
        Private ihBiases() As Double
        Private ihOutputs() As Double
        Private hoWeights()() As Double ' hidden-to-output
        Private hoSums() As Double
        Private hoBiases() As Double
        Private outputs() As Double

        Public Sub New(ByVal numInput As Integer, ByVal numHidden As Integer, ByVal numOutput As Integer)
            Me.numInput = numInput
            Me.numHidden = numHidden
            Me.numOutput = numOutput

            inputs = New Double(numInput - 1) {}
            ihWeights = Helpers.MakeMatrix(numInput, numHidden)
            ihSums = New Double(numHidden - 1) {}
            ihBiases = New Double(numHidden - 1) {}
            ihOutputs = New Double(numHidden - 1) {}
            hoWeights = Helpers.MakeMatrix(numHidden, numOutput)
            hoSums = New Double(numOutput - 1) {}
            hoBiases = New Double(numOutput - 1) {}
            outputs = New Double(numOutput - 1) {}
        End Sub

        Public Sub SetWeights(ByVal weights() As Double)
            Dim numWeights As Integer = (numInput * numHidden) + (numHidden * numOutput) + numHidden + numOutput
            If weights.Length <> numWeights Then
                Throw New Exception("The weights array length: " & weights.Length & " does not match the total number of weights and biases: " & numWeights)
            End If

            Dim k As Integer = 0 ' points into weights param

            For i = 0 To numInput - 1
                For j = 0 To numHidden - 1
                    ihWeights(i)(j) = weights(k)
                    k += 1
                Next j
            Next i

            For i = 0 To numHidden - 1
                ihBiases(i) = weights(k)
                k += 1
            Next i

            For i = 0 To numHidden - 1
                For j = 0 To numOutput - 1
                    hoWeights(i)(j) = weights(k)
                    k += 1
                Next j
            Next i

            For i = 0 To numOutput - 1
                hoBiases(i) = weights(k)
                k += 1
            Next i
        End Sub

        Public Function ComputeOutputs(ByVal xValues() As Double) As Double()
            If xValues.Length <> numInput Then
                Throw New Exception("Inputs array length " & inputs.Length & " does not match NN numInput value " & numInput)
            End If

            For i = 0 To numHidden - 1
                ihSums(i) = 0.0
            Next i
            For i = 0 To numOutput - 1
                hoSums(i) = 0.0
            Next i

            For i = 0 To xValues.Length - 1 ' copy x-values to inputs
                Me.inputs(i) = xValues(i)
            Next i

            'Console.WriteLine("Inputs:")
            'Helpers.ShowVector(this.inputs)

            'Console.WriteLine("input-to-hidden weights:")
            'Helpers.ShowMatrix(this.ihWeights, -1)

            For j = 0 To numHidden - 1 ' compute input-to-hidden weighted sums
                For i = 0 To numInput - 1
                    ihSums(j) += Me.inputs(i) * ihWeights(i)(j)
                Next i
            Next j

            'Console.WriteLine("input-to-hidden sums:")
            'Helpers.ShowVector(this.ihSums)

            'Console.WriteLine("input-to-hidden biases:")
            'Helpers.ShowVector(ihBiases)

            For i = 0 To numHidden - 1 ' add biases to input-to-hidden sums
                ihSums(i) += ihBiases(i)
            Next i

            Console.WriteLine(vbCrLf & "input-to-hidden sums after adding i-h biases:")
            Helpers.ShowVector(Me.ihSums)

            For i = 0 To numHidden - 1 ' determine input-to-hidden output
                'ihOutputsi = StepFunction(ihSumsi); // step function
                ihOutputs(i) = SigmoidFunction(ihSums(i))
            Next i

            'Console.WriteLine("\ninput-to-hidden outputs after sigmoid:")
            'Helpers.ShowVector(this.ihOutputs)

            'Console.WriteLine("hidden-to-output weights:")
            'Helpers.ShowMatrix(hoWeights, -1)

            For j = 0 To numOutput - 1 ' compute hidden-to-output weighted sums
                For i = 0 To numHidden - 1
                    hoSums(j) += ihOutputs(i) * hoWeights(i)(j)
                Next i
            Next j

            'Console.WriteLine("hidden-to-output sums:")
            'Helpers.ShowVector(hoSums)

            'Console.WriteLine("hidden-to-output biases:")
            'Helpers.ShowVector(this.hoBiases)

            For i = 0 To numOutput - 1 ' add biases to input-to-hidden sums
                hoSums(i) += hoBiases(i)
            Next i

            Console.WriteLine("hidden-to-output sums after adding h-o biases:")
            Helpers.ShowVector(Me.hoSums)

            For i = 0 To numOutput - 1 ' determine hidden-to-output result
                Me.outputs(i) = HyperTanFunction(hoSums(i))
            Next i

            Dim result(numOutput - 1) As Double
            Me.outputs.CopyTo(result, 0)

            Return result
        End Function ' ComputeOutputs

        Private Shared Function StepFunction(ByVal x As Double) As Double
            If x > 0.0 Then
                Return 1.0
            Else
                Return 0.0
            End If
        End Function

        Private Shared Function SigmoidFunction(ByVal x As Double) As Double
            If x < -45.0 Then
                Return 0.0
            ElseIf x > 45.0 Then
                Return 1.0
            Else
                Return 1.0 / (1.0 + Math.Exp(-x))
            End If
        End Function

        Private Shared Function HyperTanFunction(ByVal x As Double) As Double
            If x < -10.0 Then
                Return -1.0
            ElseIf x > 10.0 Then
                Return 1.0
            Else
                Return Math.Tanh(x)
            End If
        End Function
    End Class ' Class NeuralNetwork

    Public Module Helpers
        Public Function MakeMatrix(ByVal rows As Integer, ByVal cols As Integer) As Double()()
            Dim result(rows - 1)() As Double
            For i = 0 To rows - 1
                result(i) = New Double(cols - 1) {}
            Next i
            Return result
        End Function

        Public Sub ShowVector(ByVal vector() As Double)
            For i = 0 To vector.Length - 1
                If i > 0 AndAlso i Mod 12 = 0 Then ' max of 12 values per row
                    Console.WriteLine("")
                End If
                If vector(i) >= 0.0 Then
                    Console.Write(" ")
                End If
                Console.Write(vector(i).ToString("F2") & " ") ' 2 decimals
            Next i
            Console.WriteLine(vbCrLf)
        End Sub

        Public Sub ShowMatrix(ByVal matrix()() As Double, ByVal numRows As Integer)
            Dim ct As Integer = 0
            If numRows = -1 Then ' If numRows == -1, show all rows
                numRows = Integer.MaxValue
            End If
            Dim i As Integer = 0
            Do While i < matrix.Length AndAlso ct < numRows
                For j = 0 To matrix(0).Length - 1
                    If matrix(i)(j) >= 0.0 Then ' blank space instead of '+' sign
                        Console.Write(" ")
                    End If
                    Console.Write(matrix(i)(j).ToString("F2") & " ")
                Next j
                Console.WriteLine("")
                ct += 1
                i += 1
            Loop
            Console.WriteLine("")
        End Sub
    End Module  ' Module Helpers
End Namespace ' ns
